"""Add complete-field to Sources

Revision ID: 78fa69083a7c
Revises: 87d4ea0aa91d
Create Date: 2019-10-11 19:06:17.373389

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "78fa69083a7c"
down_revision = "87d4ea0aa91d"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("source", sa.Column("complete", sa.Boolean(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("source", "complete")
    # ### end Alembic commands ###



import os
import pickle as pkl
import re
import sys
from typing import Iterable

import requests

try:
    from antlr4 import CommonTokenStream
    from antlr4 import InputStream
except:
    SUPPORTS_ANTLR = False
else:
    SUPPORTS_ANTLR = True

import gavel.config.settings as settings
import gavel.dialects.db.structures as db
from gavel.config import settings as settings
from gavel.dialects.base.parser import LogicParser, Parseable, Target, StringBasedParser
from gavel.dialects.base.parser import ParserException
from gavel.dialects.base.parser import ProblemParser
from gavel.dialects.base.parser import ProofParser
from gavel.dialects.db.compiler import DBCompiler
from gavel.dialects.db.connection import get_or_create
from gavel.dialects.db.connection import get_or_None
from gavel.dialects.db.connection import with_session
from gavel.dialects.tptp.sources import InferenceSource
from gavel.dialects.tptp.sources import Input
from gavel.dialects.tptp.sources import InternalSource
from gavel.logic import logic
from gavel.logic import problem
from gavel.logic.logic import LogicElement
from gavel.logic.problem import AnnotatedFormula
from gavel.logic.problem import FormulaRole
from gavel.logic.problem import Problem
from gavel.logic.proof import Axiom
from gavel.logic.proof import Inference
from gavel.logic.proof import Introduction
from gavel.logic.proof import LinearProof
from gavel.logic.proof import ProofStep

if SUPPORTS_ANTLR:
    from .antlr4.flattening import FOFFlatteningVisitor
    from .antlr4.tptp_v7_0_0_0Lexer import tptp_v7_0_0_0Lexer
    from .antlr4.tptp_v7_0_0_0Parser import tptp_v7_0_0_0Parser

from lark import Lark, Tree

sys.setrecursionlimit(10000)

form_expression = re.compile(r"^(?!%|\n)(?P<logic>[^(]*)\([^.]*\)\.\S*$")

_BINARY_CONNECTIVE_MAP = {
    "&": logic.BinaryConnective.CONJUNCTION,
    "|": logic.BinaryConnective.DISJUNCTION,
    "=>": logic.BinaryConnective.IMPLICATION,
    "<=>": logic.BinaryConnective.BIIMPLICATION,
    "<=": logic.BinaryConnective.REVERSE_IMPLICATION,
    "<~>": logic.BinaryConnective.SIMILARITY,
    "~&": logic.BinaryConnective.NEGATED_CONJUNCTION,
    "~|": logic.BinaryConnective.NEGATED_DISJUNCTION,
    "=": logic.BinaryConnective.EQ,
    "!=": logic.BinaryConnective.NEQ,
    "@": logic.BinaryConnective.APPLY,
    "*": logic.BinaryConnective.PRODUCT,
    "+": logic.BinaryConnective.UNION,
    "-->": logic.BinaryConnective.GENTZEN_ARROW,
    ":=": logic.BinaryConnective.ASSIGN,
    ">": logic.BinaryConnective.ARROW,
}

with open(os.path.join(os.path.dirname(__file__), "tptp.lark")) as gf:
    lark_grammar = Lark(gf.read(), start=["start", "tptp_line"])


def _balance_binary_tree(obj, skip_connective=True, **kwargs):
    if len(obj.children) > 2:
        split_point = len(obj.children) // 2
        if skip_connective:
            cl = obj.children[:split_point]
            cr = obj.children[split_point:]
        else:
            # It the split marker is on an operator everything is fine.
            # Otherwise move it
            if not (
                isinstance(obj.children[split_point], str)
                and obj.children[split_point] in _BINARY_CONNECTIVE_MAP
            ):
                split_point -= 1
            connective = obj.children[split_point]
            cl = obj.children[:split_point]
            cr = obj.children[split_point + 1 :]

        tl = Tree(data=obj.data, children=cl)
        tr = Tree(data=obj.data, children=cr)
        l = _balance_binary_tree(tl, skip_connective=skip_connective, **kwargs)
        r = _balance_binary_tree(tr, skip_connective=skip_connective, **kwargs)
        if skip_connective:
            return Tree(data=obj.data, children=[l, r])
        else:
            return Tree(data=obj.data, children=[l, connective, r])
    else:
        return obj


def _recursive_binary(skip_connective=True):
    def wrapper(f):
        def inner(self, obj, **kwargs):
            if len(obj.children) == 1:
                return self.visit(obj.children[0], **kwargs)
            else:
                return f(
                    self,
                    _balance_binary_tree(obj, skip_connective=skip_connective),
                    **kwargs
                )

        return inner

    return wrapper


class TPTPParser(LogicParser, StringBasedParser):
    def is_valid(self, inp: str) -> bool:
        pass

    def parse(self, structure: Parseable, *args, **kwargs) -> Target:
        return self.visit(structure)

    def load_single_from_string(self, string: str, *args, **kwargs):
        r = lark_grammar.parse(string, start="tptp_line")
        return r

    def visit(self, obj: Tree, **kwargs):
        if isinstance(obj, str):
            return obj
        meth = getattr(self, "visit_%s" % obj.data, None)

        if meth is None:
            raise Exception(
                "Visitor '{name}' not found for {cls}".format(
                    name=obj.data, cls=type(obj)
                )
            )
        return meth(obj, **kwargs)

    def visit_start(self, obj, **kwargs):
        assert len(obj.children) == 1
        return self.visit(obj.children[0], **kwargs)

    def visit_tptp_line(self, obj, **kwargs):
        return self.visit(obj.children[0], **kwargs)

    def visit_annotated_formula(self, obj, **kwargs):
        return problem.AnnotatedFormula(
            logic=obj.children[0],
            name=obj.children[1],
            role=self._ROLE_MAP[obj.children[2]],
            formula=self.visit(obj.children[3], **kwargs),
        )

    def visit_formula(self, obj, **kwargs):
        return self.visit(obj.children[0], **kwargs)

    def visit_functor_term(self, obj, term_level=False, **kwargs):
        c0 = obj.children[0]
        is_defined = c0.startswith("$")
        if len(obj.children) > 1:
            if term_level:
                return logic.FunctorExpression(
                    functor=c0,
                    arguments=[
                        self.visit(c, term_level=term_level, **kwargs)
                        for c in obj.children[1:]
                    ],
                )
            else:
                p = self._DEFINED_PREDICATE_MAP[c0] if is_defined else c0
                return logic.PredicateExpression(
                    predicate=p,
                    arguments=[
                        self.visit(c, term_level=True, **kwargs)
                        for c in obj.children[1:]
                    ],
                )
        else:
            assert len(obj.children) == 1 and isinstance(obj.children[0], str)
            if is_defined:
                if c0 == "$true":
                    return logic.DefinedConstant.VERUM
                elif c0 == "$false":
                    return logic.DefinedConstant.FALSUM
                else:
                    return logic.DefinedConstant(c0)
            else:
                if c0[0] == "\"":
                    return logic.DistinctObject(c0)
                else:
                    return logic.Constant(c0)

    def visit_quantified_formula(self, obj, **kwargs):
        if len(obj.children) == 1:
            return self.visit(obj.children[0])
        else:
            q = (
                logic.Quantifier.UNIVERSAL
                if obj.children[0] == "!"
                else logic.Quantifier.EXISTENTIAL
            )
            return logic.QuantifiedFormula(
                quantifier=q,
                variables=[self.visit(c, **kwargs) for c in obj.children[1:-1]],
                formula=self.visit(obj.children[-1], **kwargs),
            )

    def visit_unary_formula(self, obj, **kwargs):
        if len(obj.children) == 1:
            return self.visit(obj.children[0])
        else:
            return logic.UnaryFormula(
                connective=logic.UnaryConnective.NEGATION,
                formula=self.visit(obj.children[1], **kwargs),
            )

    @_recursive_binary(skip_connective=True)
    def visit_conjunction(self, obj, **kwargs):
        return logic.BinaryFormula(
            left=self.visit(obj.children[0], **kwargs),
            operator=logic.BinaryConnective.CONJUNCTION,
            right=self.visit(obj.children[1], **kwargs),
        )

    @_recursive_binary(skip_connective=True)
    def visit_disjunction(self, obj, **kwargs):
        return logic.BinaryFormula(
            left=self.visit(obj.children[0], **kwargs),
            operator=logic.BinaryConnective.DISJUNCTION,
            right=self.visit(obj.children[1], **kwargs),
        )

    @_recursive_binary(skip_connective=False)
    def visit_binary_formula(self, obj, **kwargs):
        return logic.BinaryFormula(
            left=self.visit(obj.children[0], **kwargs),
            operator=self.visit_binary_operator(obj.children[1], **kwargs),
            right=self.visit(obj.children[2], **kwargs),
        )

    def visit_type_binary_formula(self, obj, **kwargs):
        return self.visit_binary_formula(obj, **kwargs)

    def visit_logic_binary_formula(self, obj, **kwargs):
        return self.visit_binary_formula(obj, **kwargs)

    def visit_object_binary_formula(self, obj, **kwargs):
        if len(obj.children) > 1:
            kwargs["term_level"] = True
        return self.visit_binary_formula(obj, **kwargs)

    _ROLE_MAP = {
        "axiom": problem.FormulaRole.AXIOM,
        "hypothesis": problem.FormulaRole.HYPOTHESIS,
        "definition": problem.FormulaRole.DEFINITION,
        "assumption": problem.FormulaRole.ASSUMPTION,
        "lemma": problem.FormulaRole.LEMMA,
        "theorem": problem.FormulaRole.THEOREM,
        "corollary": problem.FormulaRole.COROLLARY,
        "conjecture": problem.FormulaRole.CONJECTURE,
        "negated_conjecture": problem.FormulaRole.NEGATED_CONJECTURE,
        "plain": problem.FormulaRole.PLAIN,
        "type": problem.FormulaRole.TYPE,
        "fi_domain": problem.FormulaRole.FINITE_INTERPRETATION_DOMAIN,
        "fi_functors": problem.FormulaRole.FINITE_INTERPRETATION_FUNCTORS,
        "fi_predicates": problem.FormulaRole.FINITE_INTERPRETATION_PREDICATES,
        "unknown": problem.FormulaRole.UNKNOWN,
    }

    _DEFINED_PREDICATE_MAP = {
        "$distinct": logic.DefinedPredicate.DISTINCT,
        "$less": logic.DefinedPredicate.LESS,
        "$lesseq": logic.DefinedPredicate.LESS_EQ,
        "$greater": logic.DefinedPredicate.GREATER,
        "$greatereq": logic.DefinedPredicate.GREATER_EQ,
        "$is_int": logic.DefinedPredicate.IS_INT,
        "$is_rat": logic.DefinedPredicate.IS_RAT,
        "$box_P": logic.DefinedPredicate.BOX_P,
        "$box_i": logic.DefinedPredicate.BOX_I,
        "$box_int": logic.DefinedPredicate.BOX_INT,
        "$box": logic.DefinedPredicate.BOX,
        "$dia_P": logic.DefinedPredicate.DIA_P,
        "$dia_i": logic.DefinedPredicate.DIA_I,
        "$dia_int": logic.DefinedPredicate.DIA_INT,
        "$dia": logic.DefinedPredicate.DIA,
    }

    def visit_binary_operator(self, obj, **kwargs):
        return _BINARY_CONNECTIVE_MAP[obj]

    def visit_variable(self, obj, **kwargs):
        assert len(obj.children) == 1 and isinstance(obj.children[0], str)
        return logic.Variable(obj.children[0])

    def stream_formula_lines(self, lines: Iterable[str], **kwargs):
        buffer = ""
        for line in lines:
            if not line.startswith("%") and not line.startswith("\n"):
                buffer += line.strip()
                if buffer.endswith("."):
                    yield buffer
                    buffer = ""
        if buffer:
            raise Exception('Unprocessed input: """%s"""' % buffer)

    def visit_distinct_object(self, obj, **kwargs):
        assert len(obj.children) == 1 and isinstance(obj.children[0], str)
        return logic.DistinctObject(obj.children[0][1:-1])


if SUPPORTS_ANTLR:
    class TPTPAntlrParser(LogicParser, StringBasedParser):
        visitor = FOFFlatteningVisitor()

        def folder_processor(self, path, file_processor, *args, **kwargs):
            for root, dirs, files in os.walk(path):
                for file in files:
                    if not file == "README":
                        f = os.path.join(root, file)
                        yield file_processor(f, *args, **kwargs)

        def problemset_processor(self, *args, **kwargs):
            tptp_root = kwargs.get("tptp_root", settings.TPTP_ROOT)
            return self.folder_processor(
                os.path.join(tptp_root, "Problems"), self.problem_processor
            )

        def axiomsets_processor(self, *args, **kwargs):
            tptp_root = kwargs.get("tptp_root", settings.TPTP_ROOT)
            return self.folder_processor(
                os.path.join(tptp_root, "Axioms"), self.axiomset_processor
            )

        def axiomset_processor(self, path, *args, **kwargs):
            for item in self.load_expressions_from_file(path):
                yield self.formula_processor(item, *args, **kwargs)

        def formula_processor(self, formula, *args, orig=None, **kwargs):
            return formula

        def stream_formula_lines(self, lines: Iterable[str], **kwargs):
            buffer = ""
            for line in lines:
                if not line.startswith("%") and not line.startswith("\n"):
                    buffer += line.strip()
                    if buffer.endswith("."):
                        yield buffer
                        buffer = ""
            if buffer:
                raise Exception('Unprocessed input: """%s"""' % buffer)

        def load_single_from_string(self, string: str, *args, **kwargs):
            s = tptp_v7_0_0_0Parser(
                CommonTokenStream(tptp_v7_0_0_0Lexer(InputStream(string)))
            )
            return s.tptp_input()

        def load_expressions_from_file(
            self, path, *args, **kwargs
        ) -> Iterable[LogicElement]:
            with open(path) as infile:
                lines = infile.readlines()
                for line in self.load_many(lines):
                    yield self.parse(line)

        def parse(self, tree, *args, **kwargs):
            return self.visitor.visit(tree)

        def problem_processor(self, path, *args, load_imports=False, **kwargs):
            axioms = []
            imports = []
            conjectures = []
            for raw_line in self.load_expressions_from_file(path, *args, **kwargs):
                line = self.formula_processor(raw_line)
                if isinstance(line, logic.Import):
                    imports.append(line)
                elif isinstance(line, AnnotatedFormula):
                    if line.role in (
                        FormulaRole.CONJECTURE,
                        FormulaRole.NEGATED_CONJECTURE,
                    ):
                        conjectures.append(line)
                    else:
                        axioms.append(line)

            for im in imports:
                for imported_axiom in self.load_expressions_from_file(
                    os.path.join(settings.TPTP_ROOT, im.path), *args, **kwargs
                ):
                    axioms.append(imported_axiom)

            for conjecture in conjectures:
                yield Problem(premises=axioms, imports=imports, conjecture=conjecture)


class TPTPProblemParser(ProblemParser):
    logic_parser_cls = TPTPParser


class SimpleTPTPProofParser(ProofParser):
    def __init__(self):
        self._tptp_parser = TPTPParser()

    def parse(self, structure: str, *args, **kwargs):
        return LinearProof(
            steps=[
                self._create_proof_step(self._tptp_parser.parse(s))
                for s in self._tptp_parser.load_many(structure.split("\n"))
            ]
        )

    def _create_proof_step(self, e: LogicElement) -> ProofStep:
        if isinstance(e, AnnotatedFormula):
            if e.role == FormulaRole.AXIOM:
                return Axiom(formula=e.formula, name=e.name)
            elif e.role == FormulaRole.PLAIN:
                if isinstance(e.annotation, InferenceSource):
                    return Inference(
                        formula=e.formula, name=e.name, antecedents=e.annotation.parents
                    )
                elif isinstance(e.annotation, InternalSource):
                    return Introduction(
                        formula=e.formula,
                        name=e.name,
                        introduction_type=e.annotation.intro_type,
                    )
            raise ParserException(e)
        else:
            raise ParserException(e)


def get_all_files(path):
    for root, dirs, files in os.walk(path):
        for file in files:
            if not file == "README":
                yield os.path.join(root, file)


def all_problems(processor):
    for f in get_all_files(os.path.join(settings.TPTP_ROOT, "Problems")):
        print(f)
        for problem in processor.problem_processor(f):
            print("done")


def all_solution(path, system=""):
    path = os.path.join(path, "Problems")
    files = get_all_files(path)
    for file in files:
        domain, problem = os.path.normpath(file).split(os.sep)[-2:]
        response = requests.get(
            "http://www.tptp.org/cgi-bin/SeeTPTP?Category=Solutions"
            "&Domain={domain}"
            "&File={problem}"
            "&System=Vampire---4.3.THM-Ref.s".format(
                domain=domain, problem=problem[:-2]
            )
        )
        print(response)



from gavel.dialects.tptp.parser import TPTPParser
from gavel.dialects.tptp.compiler import TPTPCompiler
from gavel.logic import logic
from gavel.logic import problem
from ..test_base.test_parser import TestLogicParser, check_wrapper
from gavel.config.settings import TPTP_ROOT
import os
import unittest
import multiprocessing as mp
from unittest import skip

class TestTPTPParser(TestLogicParser):
    _parser_cls = TPTPParser

    def test_verum(self):
        inp = """fof(mElmSort,axiom,(
    ! [W0] :
      ( aElement0(W0)
     => $true ) ))."""
        result = problem.AnnotatedFormula(
            logic="fof",
            name="mElmSort",
            role=problem.FormulaRole.AXIOM,
            formula=logic.QuantifiedFormula(
                quantifier=logic.Quantifier.UNIVERSAL,
                variables=[logic.Variable("W0")],
                formula=logic.BinaryFormula(
                    left=logic.PredicateExpression(
                        predicate="aElement0", arguments=[logic.Variable("W0")]
                    ),
                    operator=logic.BinaryConnective.IMPLICATION,
                    right=logic.DefinedConstant.VERUM,
                ),
            ),
        )
        self.check_parser(inp, result)

    def test_functor(self):
        inp = """cnf(and_definition1,axiom,( and(X,n0) = n0 ))."""
        result = problem.AnnotatedFormula(
            logic="cnf",
            name="and_definition1",
            role=problem.FormulaRole.AXIOM,
            formula=logic.BinaryFormula(
                left=logic.FunctorExpression(
                    functor="and", arguments=[logic.Variable("X"), logic.Constant("n0")]
                ),
                operator=logic.BinaryConnective.EQ,
                right=logic.Constant("n0"),
            ),
        )
        self.check_parser(inp, result)

    @check_wrapper()
    def test_single_quote(self):
        inp = "p('This is arbitrary text')"
        result = logic.PredicateExpression("p",[logic.Constant("'This is arbitrary text'")])
        return inp,result

    @check_wrapper()
    def test_double_quote(self):
        inp = "p(\"This is arbitrary text\")"
        result = logic.PredicateExpression("p",[logic.DistinctObject("\"This is arbitrary text\"")])
        return inp,result

    @check_wrapper()
    def test_quantifier(self):
        inp = "![X1,X2]:?[Y1,Y2]:p(X1,X2,Y1,Y2)"
        result = logic.QuantifiedFormula(
            logic.Quantifier.UNIVERSAL,
            [logic.Variable("X1"), logic.Variable("X2")],
            logic.QuantifiedFormula(
                logic.Quantifier.EXISTENTIAL,
                [logic.Variable("Y1"), logic.Variable("Y2")],
                logic.PredicateExpression("p", [logic.Variable("X1"), logic.Variable("X2"),logic.Variable("Y1"), logic.Variable("Y2")])
            )
        )
        return inp, result

"""
class TestTHFParser(TestLogicParser):
    _parser_cls = TPTPParser

    @skip
    def test_type_formula(self):
        inp = "thf(prop_a,type,(
    prop_a: $i > $o ))."
        expected = problem.AnnotatedFormula(
            logic="thf",
            name="prop_a",
            role=problem.FormulaRole.TYPE,
            formula=logic.TypeFormula(
                name=logic.Constant("prop_a"),
                type_expression=logic.MappingType(
                    left="$i",
                    right="$o"
                )
            )
        )
        self.check_parser(inp, expected)
"""



"""Base

Revision ID: 0

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "0"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



from abc import ABC
from typing import Generic
from typing import Iterable
from typing import TypeVar

from gavel.logic.logic import LogicElement
from gavel.logic.problem import Problem
from gavel.logic.problem import Sentence
from gavel.logic.proof import Proof

Parseable = TypeVar("Parseable")
Target = TypeVar("Target")


class Parser(Generic[Parseable, Target]):
    def parse(self, structure: Parseable, *args, **kwargs) -> Target:
        """
        Transforms the input structure into metadata as used by the
        OpenEnergyPlatform

        Parameters
        ----------

        inp: str
            The input string that should be parsed into OEP metadata

        Returns
        -------

        OEPMetadata
            OEP metadata represented by `inp`
        """
        raise NotImplementedError


class StringBasedParser(Parser, ABC):
    def load_single_from_string(self, string: str, *args, **kwargs) -> Parseable:
        """
        Load a string into the structure represented by the dialect

        Parameters
        ----------
        string

        Returns
        -------
        """
        raise NotImplementedError

    def parse_single_from_string(
        self,
        string: str,
        load_args=None,
        parse_args=None,
        load_kwargs=None,
        parse_kwargs=None,
    ) -> Target:
        """
        Parse a string into OEPMetadata

        Parameters
        ----------

        string

        Returns
        -------
        """
        return self.parse(
            self.load_single_from_string(
                string, *(load_args or []), **(load_kwargs or {})
            ),
            *(parse_args or []),
            **(parse_kwargs or {})
        )

    @staticmethod
    def _unpack_file(*args, **kwargs) -> Iterable[str]:
        """
        Parameters
        ----------
        Returns
        -------
        """
        with open(*args, **kwargs) as inp:
            return inp.readlines()

    def parse_from_file(self, file_path, *args, **kwargs) -> Iterable[Target]:
        for line in self.load_many(self._unpack_file(file_path)):
            yield self.parse(line, *args, **kwargs)

    def is_valid(self, inp: str) -> bool:
        """
        Verify if `inp` is a sting representation that is parsable by this
        parser
        Parameters
        ----------
        inp: str
            String to verify
        Returns
        -------
        bool:
            Indicated whether this object is parsable or not
        """
        raise NotImplementedError

    def is_file_valid(self, *args, **kwargs):
        return self.is_valid(self.__unpack_file(*args, **kwargs))  #

    def stream_formula_lines(self, lines: Iterable[str], **kwargs):
        raise NotImplementedError

    def stream_formulas(self, path, *args, **kwargs):
        return self.stream_formula_lines(self._unpack_file(path))

    def load_many(
        self, lines: Iterable[str], *args, **kwargs
    ) -> Iterable[LogicElement]:
        return map(
            self.load_single_from_string, self.stream_formula_lines(lines, **kwargs)
        )


class LogicParser(Parser[Parseable, LogicElement]):
    pass


class ProblemParser(Parser[Parseable, Problem]):
    logic_parser_cls = LogicParser

    def __init__(self, *args, **kwargs):
        self.logic_parser = self.logic_parser_cls(*args, **kwargs)

    def parse(self, inp, *args, **kwargs):
        premises = []
        conjectures = []
        for s in map(self.logic_parser.parse_single_from_string, self.logic_parser.stream_formula_lines(inp)):
            if isinstance(s, Sentence):
                if s.is_conjecture():
                    conjectures.append(s)
                else:
                    premises.append(s)
            else:
                raise ParserException("Unknown element:" + str(s))
        for c in conjectures:
            yield Problem(premises, c)


class ParserException(Exception):
    pass


class ProofParser(Parser[Parseable, Proof]):
    pass



"""
Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later, but that will cause
  problems: the code will get executed twice:

  - When you run `python -mgavel` python will execute
    ``__main__.py`` as a script. That means there won't be any
    ``gavel.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there's no ``gavel.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""

import click
import os

import gavel.config.settings as settings
from gavel.dialects.db.structures import (
    store_formula,
    mark_source_complete,
    is_source_complete,
    store_all,
)
import gavel.dialects.tptp.parser as build_tptp
from gavel.dialects.tptp.compiler import TPTPCompiler
from gavel.dialects.tptp.parser import TPTPProblemParser
from gavel.dialects.db.compiler import DBCompiler
from gavel.dialects.tptp.parser import Problem
from gavel.dialects.tptp.parser import TPTPParser
from gavel.prover.hets.interface import HetsProve
from gavel.prover.registry import get_prover
from gavel.selection.selector import Sine
from alembic import command
from alembic.config import Config

ROOT_DIR = os.path.dirname(__file__)

alembic_cfg = Config(os.path.join(ROOT_DIR, "alembic.ini"))
alembic_cfg.set_main_option("script_location", os.path.join(ROOT_DIR, "alembic"))

@click.group()
def db():
    pass


@click.command()
def migrate_db():
    """Create tables for storage of formulas"""
    command.upgrade(alembic_cfg, "head")


@click.command()
@click.argument("path", default=None)
@click.option("-r", default=False)
def store(path, r):
    parser = TPTPParser()
    compiler = DBCompiler()
    store_all(path, parser, compiler)


@click.command()
@click.option("-p", default=settings.TPTP_ROOT)
def store_problems(p):
    settings.TPTP_ROOT = p
    build_tptp.store_problems()


@click.command()
@click.option("-p", default=settings.TPTP_ROOT)
def store_solutions(p):
    """Drop tables created gy init-db"""
    build_tptp.all_solution(p)


@click.command()
def drop_db():
    """Drop tables created gy init-db"""
    command.downgrade(alembic_cfg, "base")


@click.command()
@click.option("-p", default=settings.TPTP_ROOT)
def clear_db(p):
    """Drop tables created gy init-db and recreate them"""
    command.downgrade(alembic_cfg, "base")
    command.upgrade(alembic_cfg, "head")


@click.command()
@click.argument("p")
@click.argument("f")
@click.option("-s", default=None)
@click.option("--hets", is_flag=True, default=False)
@click.option("--plot", is_flag=True, default=False)
def prove(p, f, s, plot, hets):
    prover_interface = get_prover(p)
    if hets:
        prover_interface = HetsProve(prover_interface)
    processor = TPTPProblemParser()
    problems = list(processor.parse(f))
    compiler = TPTPCompiler()
    for problem in problems:
        if s is not None:
            selector = Sine(
                premises=problem.premises, conjecture=problem.conjecture, max_depth=10
            )
            problem = Problem(premises=selector.select(), conjecture=problem.conjecture)
        proof = prover_interface.prove(problem, compiler)
        if not plot:
            for s in proof.steps:
                print(
                    "{name}: {formula} [{source}]".format(
                        name=s.name, formula=s.formula, source=s.render_source()
                    )
                )
        else:
            g = proof.get_graph()
            g.render()


@click.command()
@click.argument("f")
def select(f):
    processor = TPTPParser()
    problem = list(processor.problem_processor(f))[0]
    selector = Sine(
        premises=problem.premises, conjecture=problem.conjecture, max_depth=10
    )
    smaller_problem = Problem(premises=selector.select(), conjecture=problem.conjecture)
    for prem in list(smaller_problem.premises):
        print(prem)
    print(smaller_problem.conjecture)


db.add_command(migrate_db)
db.add_command(drop_db)
db.add_command(clear_db)
db.add_command(store_problems)
db.add_command(store)
db.add_command(store_solutions)

db.add_command(prove)

db.add_command(select)

cli = click.CommandCollection(sources=[db])

main = cli

if __name__ == "__main__":
    cli()



"""Initial revision

Revision ID: 87d4ea0aa91d
Revises: 
Create Date: 2019-10-11 18:30:19.666242

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "87d4ea0aa91d"
down_revision = "0"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "source",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("path", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("path"),
    )
    op.create_table(
        "formula",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.Text(), nullable=True),
        sa.Column("source_id", sa.Integer(), nullable=False),
        sa.Column("json", sa.JSON(), nullable=True),
        sa.ForeignKeyConstraint(["source_id"], ["source.id"]),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "problem",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("source_id", sa.Integer(), nullable=False),
        sa.Column("conjecture_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(["conjecture_id"], ["formula.id"]),
        sa.ForeignKeyConstraint(["source_id"], ["source.id"]),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "association",
        sa.Column("left_id", sa.Integer(), nullable=False),
        sa.Column("right_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(["left_id"], ["problem.id"]),
        sa.ForeignKeyConstraint(["right_id"], ["formula.id"]),
    )
    op.create_table(
        "solution",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("problem_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(["problem_id"], ["problem.id"]),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "solution_item",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("solution_id", sa.Integer(), nullable=False),
        sa.Column("premise_id", sa.Integer(), nullable=False),
        sa.Column("used", sa.Boolean(), nullable=True),
        sa.ForeignKeyConstraint(["premise_id"], ["formula.id"]),
        sa.ForeignKeyConstraint(["solution_id"], ["solution.id"]),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("solution_item")
    op.drop_table("solution")
    op.drop_table("association")
    op.drop_table("problem")
    op.drop_table("formula")
    op.drop_table("source")
    # ### end Alembic commands ###



import unittest
from itertools import chain

from gavel.dialects.base.parser import LogicElement
from gavel.dialects.base.parser import LogicParser
from gavel.logic.problem import AnnotatedFormula, FormulaRole
from gavel.logic import logic
class TestLogicParser(unittest.TestCase):
    _parser_cls = LogicParser

    def __init__(self, *args, **kwargs):
        super(TestLogicParser, self).__init__(*args, **kwargs)
        self.parser = self._parser_cls()

    def assertObjectEqual(self, result, expected):
        self.assertTrue(isinstance(result, type(expected)) or isinstance(expected, type(result)), "Types do not math: (expected: %s, got: %s)" % (type(expected), type(result)))
        if isinstance(result, LogicElement) or isinstance(result, AnnotatedFormula):
            for n in chain(result.__dict__.keys(), expected.__dict__.keys()):
                self.assertObjectEqual(getattr(result, n), getattr(expected, n))
        elif isinstance(result, list):
            for po1, po2 in zip(result, expected):
                self.assertObjectEqual(po1, po2)
        elif isinstance(result, str):
            self.assertEqual(str(result), str(expected))
        else:
            self.assertEqual(result, expected)

    def check_parser(self, parser_input, expected):
        r = self.parser.parse_single_from_string(parser_input)
        self.assertObjectEqual(r, expected)


def check_wrapper(logic_name="fof",name="name"):
    def outer(f):
        def inner(self: TestLogicParser):
            formula, expected_formula = f(self)
            string = "{logic}({name},{role},{formula}).".format(
                logic=logic_name,
                name=name,
                role="plain",
                formula=formula
            )
            result = self.parser.parse_single_from_string(string)
            expected = AnnotatedFormula(
                logic=logic_name,
                name=name,
                role=FormulaRole.PLAIN,
                formula=expected_formula
            )
            self.assertObjectEqual(result, expected)
        return inner
    return outer



import cProfile
import pstats
import unittest
from os.path import join

import pytest

import gavel.dialects.db.structures as fol_db
from gavel.config.settings import TPTP_ROOT
from gavel.dialects.db.connection import with_session
from gavel.dialects.tptp.compiler import TPTPCompiler
from gavel.dialects.tptp.parser import TPTPParser


class TestProcessor(TPTPParser):
    compiler = TPTPCompiler()

    def parse(self, tree, *args, **kwargs):
        original = tree.getText()
        internal = self.visitor.visit(tree)
        if internal.logic != "thf":
            reconstructed = self.compiler.visit(internal)
            assert original.replace("(", "").replace(")", "") == reconstructed.replace(
                "(", ""
            ).replace(")", ""), (original, reconstructed)
            print(reconstructed)
        return internal

axioms = ["GRP001-0.ax"]

problems = ["ALG/ALG001-1.p", "NUN/NUN030^1.p"]


class DBTest(unittest.TestCase):
    @classmethod
    def setup_class(cls):
        fol_db.create_tables()

    @classmethod
    def teardown_class(cls):
        fol_db.drop_tables()





import os

TPTP_ROOT = os.environ.get("TPTP_ROOT", "")

DBMS = os.environ.get("GAVEL_DBMS", "sqlite")

DB_CONNECTION = dict(
    user=os.environ.get("GAVEL_DB_USER", ""),
    password=os.environ.get("GAVEL_DB_PASSWORD", ""),
    host=os.environ.get("GAVEL_DB_HOST", ""),
    port=os.environ.get("GABEL_DB_PORT", ""),
    database=os.environ.get("GAVEL_DB_NAME", ":memory:"),
)

HETS_HOST = os.environ.get("HETS_HOST", "localhost")
HETS_PORT = os.environ.get("HETS_PORT", 8000)



import sqlalchemy as sqla
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relation
from sqlalchemy.orm import relationship
from gavel.dialects.db.connection import with_session, get_or_create, get_or_None
from gavel.logic.problem import AnnotatedFormula

import os
import multiprocessing as mp

Base = declarative_base()


class Source(Base):
    __tablename__ = "source"
    id = sqla.Column(sqla.Integer, primary_key=True)
    path = sqla.Column(sqla.String, unique=True)
    complete = sqla.Column(sqla.Boolean, default=False)


class Formula(Base):
    __tablename__ = "formula"
    id = sqla.Column(sqla.Integer, primary_key=True)
    name = sqla.Column(sqla.Text)
    source_id = sqla.Column(sqla.Integer, sqla.ForeignKey(Source.id), nullable=False)
    source = relationship(Source)
    logic = sqla.VARCHAR(4)
    json = sqla.Column(sqla.JSON)


association_premises = sqla.Table(
    "association",
    Base.metadata,
    sqla.Column("left_id", sqla.Integer, sqla.ForeignKey("problem.id"), nullable=False),
    sqla.Column(
        "right_id", sqla.Integer, sqla.ForeignKey("formula.id"), nullable=False
    ),
)


class SolutionItem(Base):
    id = sqla.Column(sqla.Integer, primary_key=True)
    __tablename__ = "solution_item"
    solution_id = sqla.Column(
        sqla.Integer, sqla.ForeignKey("solution.id"), nullable=False
    )
    solution = relationship("Solution")
    premise_id = sqla.Column(sqla.Integer, sqla.ForeignKey(Formula.id), nullable=False)
    premise = relationship(Formula)
    used = sqla.Column(sqla.Boolean)


class Problem(Base):
    __tablename__ = "problem"
    id = sqla.Column(sqla.Integer, primary_key=True)
    source_id = sqla.Column(sqla.Integer, sqla.ForeignKey(Source.id), nullable=False)
    source = relationship(Source)
    conjecture_id = sqla.Column(
        sqla.Integer, sqla.ForeignKey(Formula.id), nullable=False
    )
    conjecture = relationship(Formula)
    premises = relation(Formula, secondary=association_premises)
    solutions = relation("Solution")

    def create_problem_file(self, file):
        for premise in self.premises:
            file.write(premise.original)


class Solution(Base):
    __tablename__ = "solution"
    id = sqla.Column(sqla.Integer, primary_key=True)
    problem_id = sqla.Column(sqla.Integer, sqla.ForeignKey("problem.id"))
    problem = relationship(Problem)
    premises = relation(SolutionItem)


@with_session
def store_formula(
    source_name,
    struc: AnnotatedFormula,
    session=None,
    source=None,
    skip_existence_check=False,
):
    created = skip_existence_check
    structure = None

    if source is None:
        source, created = get_or_create(session, Source, path=source_name)
    # If the source object was already in the database, the formula might
    # already be present, too. Check that before storing a second copy
    if not created and not skip_existence_check:
        structure = get_or_None(session, Formula, name=struc.name, source=source)
    if structure is None:
        struc.source = source
        session.add(struc)
        return True
    else:
        return False


def store_all(path, parser, compiler):
    if os.path.isdir(path):
        for sub_path in os.listdir(path):
            sub_path = os.path.join(path, sub_path)
            if os.path.isfile(sub_path):
                store_file(sub_path, parser, compiler)
    elif os.path.isfile(path):
        store_file(path, parser, compiler)


@with_session
def store_file(path, parser, compiler, session=None):
    skip = False
    skip_reason = None
    print(path)
    fname = os.path.basename(path)
    if "=" not in fname and "^" not in fname and "_" not in fname:
        source, created = get_or_create(session, Source, path=path)
        if not source.complete:
            i = 0
            with mp.Pool(mp.cpu_count() - 1) as pool:
                for struc in pool.imap(
                    parser.parse_single_from_string, parser.stream_formulas(path)
                ):
                    i += 1
                    store_formula(path, compiler.visit(struc), session=session, source=source, skip_existence_check=created)
                mark_source_complete(path, session=session)
                print("%d formulas extracted" % i)
                print("commit to database")
                print("--- done ---")
                session.commit()

        else:
            skip = True
            skip_reason = "Already complete"
    else:
        skip = True
        skip_reason = "Not supported"
    if skip:
        print("--- Skipping - Reason: %s ---" % skip_reason)


@with_session
def mark_source_complete(source, session=None):
    session.query(Source).filter_by(path=source).update({"complete": True})
    session.commit()


@with_session
def is_source_complete(source, session=None):
    source_obj = get_or_None(session, Source, path=source)
    if source_obj is None:
        return False
    return source_obj.complete
